@@ -7,8 +7,9 @@
 set -Eeuo pipefail
 
 # Configuration:
-HACKAGE_INDEX_STATE="2020-12-21T14:48:20Z"
-MIN_HAPPY_VERSION="1.20"
+# N.B. You may want to also update the index-state in hadrian/cabal.project.
+HACKAGE_INDEX_STATE="2025-01-08T09:21:31Z"
+HAPPY_CONSTRAINT=">=1.20&&<1.21"
 MIN_ALEX_VERSION="3.2.6"
 
 TOP="$(pwd)"
@@ -370,7 +371,7 @@ function setup_toolchain() {
   esac
 
   info "Building happy..."
-  $cabal_install happy --constraint="happy>=$MIN_HAPPY_VERSION"
+  $cabal_install happy --constraint="happy$HAPPY_CONSTRAINT"
 
   info "Building alex..."
   $cabal_install alex --constraint="alex>=$MIN_ALEX_VERSION"
@@ -577,6 +578,8 @@ function install_bindist() {
           --prefix="$instdir" \
           "${args[@]+"${args[@]}"}"
       make_install_destdir "$TOP"/destdir "$instdir"
+      # And check the `--info` of the installed compiler, sometimes useful in CI log.
+      "$instdir"/bin/ghc --info
       ;;
   esac
   popd
@@ -798,6 +801,12 @@ function shell() {
   run "$cmd"
 }
 
+# Evaluate some shell code in the CI environment
+function eval_cmd() {
+  local cmd="${@: 1}"
+  eval "$cmd"
+}
+
 function lint_author(){
   base=$1
   head=$2
@@ -927,5 +936,6 @@ case $1 in
   clean) clean ;;
   save_cache) save_cache ;;
   shell) shift; shell "$@" ;;
+  eval) shift; eval_cmd "$@" ;;
   *) fail "unknown mode $1" ;;
 esac
