@@ -18,6 +18,8 @@ module GHC.Stg.Pipeline
 
 import GHC.Prelude
 
+import GHC.Driver.Flags
+
 import GHC.Stg.Syntax
 
 import GHC.Stg.Lint     ( lintStgTopBindings )
@@ -28,10 +30,9 @@ import GHC.Stg.BcPrep   ( bcPrep )
 import GHC.Stg.CSE      ( stgCse )
 import GHC.Stg.Lift     ( StgLiftConfig, stgLiftLams )
 import GHC.Unit.Module ( Module )
-import GHC.Runtime.Context ( InteractiveContext )
 
-import GHC.Driver.Flags (DumpFlag(..))
 import GHC.Utils.Error
+import GHC.Types.Var
 import GHC.Types.Unique.Supply
 import GHC.Utils.Outputable
 import GHC.Utils.Logger
@@ -50,6 +51,7 @@ data StgPipelineOpts = StgPipelineOpts
   -- ^ Should we lint the STG at various stages of the pipeline?
   , stgPipeline_pprOpts     :: !StgPprOpts
   , stgPlatform             :: !Platform
+  , stgPipeline_forBytecode :: !Bool
   }
 
 newtype StgM a = StgM { _unStgM :: ReaderT Char IO a }
@@ -60,21 +62,21 @@ newtype StgM a = StgM { _unStgM :: ReaderT Char IO a }
 type StgCgInfos = NameEnv TagSig
 
 instance MonadUnique StgM where
-  getUniqueSupplyM = StgM $ do { mask <- ask
-                               ; liftIO $! mkSplitUniqSupply mask}
-  getUniqueM = StgM $ do { mask <- ask
-                         ; liftIO $! uniqFromMask mask}
+  getUniqueSupplyM = StgM $ do { tag <- ask
+                               ; liftIO $! mkSplitUniqSupply tag}
+  getUniqueM = StgM $ do { tag <- ask
+                         ; liftIO $! uniqFromTag tag}
 
 runStgM :: Char -> StgM a -> IO a
 runStgM mask (StgM m) = runReaderT m mask
 
 stg2stg :: Logger
-        -> InteractiveContext
+        -> [Var]                     -- ^ extra vars in scope from GHCi
         -> StgPipelineOpts
-        -> Module                    -- module being compiled
-        -> [StgTopBinding]           -- input program
+        -> Module                    -- ^ module being compiled
+        -> [StgTopBinding]           -- ^ input program
         -> IO ([CgStgTopBinding], StgCgInfos) -- output program
-stg2stg logger ictxt opts this_mod binds
+stg2stg logger extra_vars opts this_mod binds
   = do  { dump_when Opt_D_dump_stg_from_core "Initial STG:" binds
         ; showPass logger "Stg2Stg"
         -- Do the main business!
@@ -93,7 +95,7 @@ stg2stg logger ictxt opts this_mod binds
           -- annotations (which is used by code generator to compute offsets into closures)
         ; let binds_sorted_with_fvs = depSortWithAnnotStgPgm this_mod binds'
         -- See Note [Tag inference for interactive contexts]
-        ; inferTags (stgPipeline_pprOpts opts) logger this_mod binds_sorted_with_fvs
+        ; inferTags (stgPipeline_pprOpts opts) (stgPipeline_forBytecode opts) logger this_mod binds_sorted_with_fvs
    }
 
   where
@@ -102,9 +104,9 @@ stg2stg logger ictxt opts this_mod binds
       = lintStgTopBindings
           (stgPlatform opts) logger
           diag_opts ppr_opts
-          ictxt this_mod unarised
+          extra_vars this_mod unarised
       | otherwise
-      = \ _whodunnit _binds -> return ()
+      = \ _whodunit _binds -> return ()
 
     -------------------------------------------
     do_stg_pass :: Module -> [StgTopBinding] -> StgToDo -> StgM [StgTopBinding]
