@@ -7,7 +7,7 @@
  *
  * ---------------------------------------------------------------------------*/
 
-#include "PosixSource.h"
+#include "rts/PosixSource.h"
 #include "Rts.h"
 
 #include "RtsUtils.h"
@@ -119,7 +119,7 @@ static bool read_heap_profiling_flag(const char *arg);
 static void read_trace_flags(const char *arg);
 #endif
 
-static void errorUsage (void) GNU_ATTRIBUTE(__noreturn__);
+static void errorUsage (void) STG_NORETURN;
 
 #if defined(mingw32_HOST_OS)
 static char** win32_full_utf8_argv;
@@ -199,11 +199,13 @@ void initRtsFlagsDefaults(void)
     RtsFlags.DebugFlags.prof            = false;
     RtsFlags.DebugFlags.apply           = false;
     RtsFlags.DebugFlags.linker          = false;
+    RtsFlags.DebugFlags.linker_verbose  = false;
     RtsFlags.DebugFlags.squeeze         = false;
     RtsFlags.DebugFlags.hpc             = false;
     RtsFlags.DebugFlags.sparks          = false;
     RtsFlags.DebugFlags.numa            = false;
     RtsFlags.DebugFlags.compact         = false;
+    RtsFlags.DebugFlags.continuation    = false;
 
 #if defined(PROFILING)
     RtsFlags.CcFlags.doCostCentres      = COST_CENTRES_NONE;
@@ -242,7 +244,8 @@ void initRtsFlagsDefaults(void)
     RtsFlags.TraceFlags.nullWriter = false;
 #endif
 
-#if defined(PROFILING)
+// See Note [No timer on wasm32]
+#if defined(PROFILING) && !defined(wasm32_HOST_ARCH)
     // When profiling we want a lot more ticks
     RtsFlags.MiscFlags.tickInterval     = USToTime(1000);  // 1ms
 #else
@@ -297,15 +300,17 @@ usage_text[] = {
 "",
 "Usage: <prog> <args> [+RTS <rtsopts> | -RTS <args>] ... --RTS <args>",
 "",
-"   +RTS    Indicates run time system options follow",
-"   -RTS    Indicates program arguments follow",
-"  --RTS    Indicates that ALL subsequent arguments will be given to the",
-"           program (including any of these RTS flags)",
+"   +RTS     Indicates run time system options follow",
+"   -RTS     Indicates program arguments follow",
+"  --RTS     Indicates that ALL subsequent arguments will be given to the",
+"            program (including any of these RTS flags)",
 "",
-"The following run time system options are available:",
+"The following run time system options may be available (note that some",
+"of these may not be usable unless this program was linked with the -rtsopts",
+"flag):",
 "",
-"  -?       Prints this message and exits; the program is not executed",
-"  --info   Print information about the RTS used by this program",
+"  -?        Prints this message and exits; the program is not executed",
+"  --info    Print information about the RTS used by this program",
 "",
 "  --nonmoving-gc",
 "            Selects the non-moving mark-and-sweep garbage collector to",
@@ -314,12 +319,12 @@ usage_text[] = {
 "            Selects the copying garbage collector to manage all generations.",
 "",
 "  -K<size>  Sets the maximum stack size (default: 80% of the heap)",
-"            Egs: -K32k -K512k -K8M",
-"  -ki<size> Sets the initial thread stack size (default 1k)  Egs: -ki4k -ki2m",
+"            e.g.: -K32k -K512k -K8M",
+"  -ki<size> Sets the initial thread stack size (default 1k)  e.g.: -ki4k -ki2m",
 "  -kc<size> Sets the stack chunk size (default 32k)",
 "  -kb<size> Sets the stack chunk buffer size (default 1k)",
 "",
-"  -A<size>  Sets the minimum allocation area size (default 4m) Egs: -A20m -A10k",
+"  -A<size>  Sets the minimum allocation area size (default 4m) e.g.: -A20m -A10k",
 "  -AL<size> Sets the amount of large-object memory that can be allocated",
 "            before a GC is triggered (default: the value of -A)",
 "  -F<n>     Sets the collecting threshold for old generations as a factor of",
@@ -333,8 +338,8 @@ usage_text[] = {
 "            (default 4.0)",
 "  -n<size>  Allocation area chunk size (0 = disabled, default: 0)",
 "  -O<size>  Sets the minimum size of the old generation (default 1M)",
-"  -M<size>  Sets the maximum heap size (default unlimited)  Egs: -M256k -M1G",
-"  -H<size>  Sets the minimum heap size (default 0M)   Egs: -H24m  -H1G",
+"  -M<size>  Sets the maximum heap size (default unlimited)  e.g.: -M256k -M1G",
+"  -H<size>  Sets the minimum heap size (default 0M)   e.g.: -H24m  -H1G",
 "  -xb<addr> Sets the address from which a suitable start for the heap memory",
 "            will be searched from. This is useful if the default address",
 "            clashes with some third-party library.",
@@ -342,13 +347,13 @@ usage_text[] = {
 "  -m<n>     Minimum % of heap which must be available (default 3%)",
 "  -G<n>     Number of generations (default: 2)",
 "  -c<n>     Use in-place compaction instead of copying in the oldest generation",
-"           when live data is at least <n>% of the maximum heap size set with",
-"           -M (default: 30%)",
-"  -c       Use in-place compaction for all oldest generation collections",
-"           (the default is to use copying)",
-"  -w       Use mark-region for the oldest generation (experimental)",
+"            when live data is at least <n>% of the maximum heap size set with",
+"            -M (default: 30%)",
+"  -c        Use in-place compaction for all oldest generation collections",
+"            (the default is to use copying)",
+"  -w        Use mark-region for the oldest generation (experimental)",
 #if defined(THREADED_RTS)
-"  -I<sec>  Perform full GC after <sec> idle time (default: 0.3, 0 == off)",
+"  -I<sec>   Perform full GC after <sec> idle time (default: 0.3, 0 == off)",
 #endif
 "",
 "  -T         Collect GC statistics (useful for in-program statistics access)",
@@ -375,6 +380,7 @@ usage_text[] = {
 "                 T = closure type",
 "                 d = closure description",
 "                 y = type description",
+"                 i = info table",
 "                 r = retainer",
 "                 b = biography (LAG,DRAG,VOID,USE)",
 "  A subset of closures may be selected thusly:",
@@ -397,13 +403,15 @@ usage_text[] = {
 #else /* PROFILING */
 "  -h       Heap residency profile (output file <program>.hp)",
 "  -hT      Produce a heap profile grouped by closure type",
+"  -hi      Produce a heap profile grouped by info table address",
 "  -po<file>  Override profiling output file name prefix (program name by default)",
 #endif /* PROFILING */
 
 "  -i<sec>  Time between heap profile samples (seconds, default: 0.1)",
-"  --no-automatic-heap-samples  Do not start the heap profile interval timer on start-up,",
-"                               Rather, the application will be responsible for triggering",
-"                               heap profiler samples."
+"  --no-automatic-heap-samples",
+"           Do not start the heap profile interval timer on start-up,",
+"           Rather, the application will be responsible for triggering",
+"           heap profiler samples."
 
 #if defined(TRACING)
 "",
@@ -428,7 +436,8 @@ usage_text[] = {
 #  endif
 "               -x    disable an event class, for any flag above",
 "             the initial enabled event classes are 'sgpu'",
-" --eventlog-flush-interval=<secs> Flush the eventlog at the specified interval.",
+" --eventlog-flush-interval=<secs>",
+"             Periodically flush the eventlog at the specified interval.",
 #endif
 
 "",
@@ -462,11 +471,13 @@ usage_text[] = {
 "  -Dp  DEBUG: prof",
 "  -Da  DEBUG: apply",
 "  -Dl  DEBUG: linker",
+"  -DL  DEBUG: linker (verbose)",
 "  -Dm  DEBUG: stm",
 "  -Dz  DEBUG: stack squeezing",
 "  -Dc  DEBUG: program coverage",
 "  -Dr  DEBUG: sparks",
 "  -DC  DEBUG: compact",
+"  -Dk  DEBUG: continuation",
 "",
 "     NOTE: DEBUG events are sent to stderr by default; add -l to create a",
 "     binary event log file instead.",
@@ -476,66 +487,66 @@ usage_text[] = {
 "  -N[<n>]    Use <n> processors (default: 1, -N alone determines",
 "             the number of processors to use automatically)",
 "  -maxN[<n>] Use up to <n> processors automatically",
-"  -qg[<n>]  Use parallel GC only for generations >= <n>",
-"            (default: 0, -qg alone turns off parallel GC)",
-"  -qb[<n>]  Use load-balancing in the parallel GC only for generations >= <n>",
-"            (default: 1 for -A < 32M, 0 otherwise;",
-"             -qb alone turns off load-balancing)",
-"  -qn<n>    Use <n> threads for parallel GC (defaults to value of -N)",
-"  -qa       Use the OS to set thread affinity (experimental)",
-"  -qm       Don't automatically migrate threads between CPUs",
-"  -qi<n>    If a processor has been idle for the last <n> GCs, do not",
-"            wake it up for a non-load-balancing parallel GC.",
-"            (0 disables,  default: 0)",
+"  -qg[<n>]   Use parallel GC only for generations >= <n>",
+"             (default: 0, -qg alone turns off parallel GC)",
+"  -qb[<n>]   Use load-balancing in the parallel GC only for generations >= <n>",
+"             (default: 1 for -A < 32M, 0 otherwise;",
+"              -qb alone turns off load-balancing)",
+"  -qn<n>     Use <n> threads for parallel GC (defaults to value of -N)",
+"  -qa        Use the OS to set thread affinity (experimental)",
+"  -qm        Don't automatically migrate threads between CPUs",
+"  -qi<n>     If a processor has been idle for the last <n> GCs, do not",
+"             wake it up for a non-load-balancing parallel GC.",
+"             (0 disables,  default: 0)",
 "  --numa[=<node_mask>]",
-"            Use NUMA, nodes given by <node_mask> (default: off)",
+"             Use NUMA, nodes given by <node_mask> (default: off)",
 #if defined(DEBUG)
 "  --debug-numa[=<num_nodes>]",
-"            Pretend NUMA: like --numa, but without the system calls.",
-"            Can be used on non-NUMA systems for debugging.",
+"             Pretend NUMA: like --numa, but without the system calls.",
+"             Can be used on non-NUMA systems for debugging.",
 "",
 #endif
 #endif
 "  --install-signal-handlers=<yes|no>",
-"            Install signal handlers (default: yes)",
+"             Install signal handlers (default: yes)",
 #if defined(mingw32_HOST_OS)
 "  --install-seh-handlers=<yes|no>",
-"            Install exception handlers (default: yes)",
+"             Install exception handlers (default: yes)",
 "  --generate-crash-dumps",
-"            Generate Windows crash dumps, requires exception handlers",
-"            to be installed. Implies --install-signal-handlers=yes.",
-"            (default: no)",
+"             Generate Windows crash dumps, requires exception handlers",
+"             to be installed. Implies --install-signal-handlers=yes.",
+"             (default: no)",
 "  --generate-stack-traces=<yes|no>",
-"            Generate a stack trace when your application encounters a",
-"            fatal error. When symbols are available an attempt will be",
-"            made to resolve addresses to names. (default: yes)",
+"             Generate a stack trace when your application encounters a",
+"             fatal error. When symbols are available an attempt will be",
+"             made to resolve addresses to names. (default: yes)",
 #endif
 "  --io-manager=<native|posix>",
-"            The I/O manager subsystem to use. (default: posix)",
+"             The I/O manager subsystem to use. (default: posix)",
 #if defined(THREADED_RTS)
 #if defined(mingw32_HOST_OS)
 "  --io-manager-threads=<num>",
-"            The number of worker threads to use in the native I/O manager to",
-"            handle completion events. (default: num cores)",
+"             The number of worker threads to use in the native I/O manager to",
+"             handle completion events. (default: num cores)",
 #endif
-"  -e<n>     Maximum number of outstanding local sparks (default: 4096)",
+"  -e<n>      Maximum number of outstanding local sparks (default: 4096)",
 #endif
 #if defined(x86_64_HOST_ARCH)
 #if !DEFAULT_LINKER_ALWAYS_PIC
-"  -xp       Assume that all object files were compiled with -fPIC",
-"            -fexternal-dynamic-refs and load them anywhere in the address",
-"            space",
+"  -xp        Assume that all object files were compiled with -fPIC",
+"             -fexternal-dynamic-refs and load them anywhere in the address",
+"             space",
 #endif
-"  -xm       Base address to mmap memory in the GHCi linker",
-"            (hex; must be <80000000)",
+"  -xm        Base address to mmap memory in the GHCi linker",
+"             (hex; must be <80000000)",
 #endif
-"  -xq       The allocation limit given to a thread after it receives",
-"            an AllocationLimitExceeded exception. (default: 100k)",
+"  -xq        The allocation limit given to a thread after it receives",
+"             an AllocationLimitExceeded exception. (default: 100k)",
 "",
 "  -Mgrace=<n>",
-"            The amount of allocation after the program receives a",
-"            HeapOverflow exception before the exception is thrown again, if",
-"            the program is still exceeding the heap limit.",
+"             The amount of allocation after the program receives a",
+"             HeapOverflow exception before the exception is thrown again, if",
+"             the program is still exceeding the heap limit.",
 "",
 "RTS options may also be specified using the GHCRTS environment variable.",
 "",
@@ -547,7 +558,7 @@ usage_text[] = {
 
 /**
 Note [Windows Unicode Arguments]
-~~~~~~~~~~~~~~~~~~~~~~~~~~
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 On Windows argv is usually encoded in the current Codepage which might not
 support unicode.
 
@@ -797,7 +808,7 @@ void setupRtsFlags (int *argc, char *argv[], RtsConfig rts_config)
  * procRtsOpts: Process rts_argv between rts_argc0 and rts_argc.
  * -------------------------------------------------------------------------- */
 
-#if defined(HAVE_UNISTD_H) && defined(HAVE_SYS_TYPES_H) && !defined(mingw32_HOST_OS)
+#if defined(HAVE_UNISTD_H) && defined(HAVE_SYS_TYPES_H) && !defined(mingw32_HOST_OS) && defined(HAVE_GETUID)
 static void checkSuid(RtsOptsEnabledEnum enabled)
 {
     if (enabled == RtsOptsSafeOnly) {
@@ -1441,7 +1452,7 @@ error = true;
 #if !defined(PROFILING)
                 switch (rts_argv[arg][2]) {
                   case '\0':
-                    errorBelch("-h is deprecated and will be removed in the next release (9.4),\nuse -hT instead.");
+                    errorBelch("-h is deprecated, use -hT instead.");
 
                     FALLTHROUGH;
                   case 'T':
@@ -1845,7 +1856,7 @@ static void normaliseRtsOpts (void)
                     RtsFlags.MiscFlags.tickInterval);
     }
 
-    if (RtsFlags.ConcFlags.ctxtSwitchTime > 0) {
+    if (RtsFlags.ConcFlags.ctxtSwitchTime > 0 && RtsFlags.MiscFlags.tickInterval != 0) {
         RtsFlags.ConcFlags.ctxtSwitchTicks =
             RtsFlags.ConcFlags.ctxtSwitchTime /
             RtsFlags.MiscFlags.tickInterval;
@@ -1853,7 +1864,7 @@ static void normaliseRtsOpts (void)
         RtsFlags.ConcFlags.ctxtSwitchTicks = 0;
     }
 
-    if (RtsFlags.ProfFlags.heapProfileInterval > 0) {
+    if (RtsFlags.ProfFlags.heapProfileInterval > 0 && RtsFlags.MiscFlags.tickInterval != 0) {
         RtsFlags.ProfFlags.heapProfileIntervalTicks =
             RtsFlags.ProfFlags.heapProfileInterval /
             RtsFlags.MiscFlags.tickInterval;
@@ -1861,7 +1872,7 @@ static void normaliseRtsOpts (void)
         RtsFlags.ProfFlags.heapProfileIntervalTicks = 0;
     }
 
-    if (RtsFlags.TraceFlags.eventlogFlushTime > 0) {
+    if (RtsFlags.TraceFlags.eventlogFlushTime > 0 && RtsFlags.MiscFlags.tickInterval != 0) {
         RtsFlags.TraceFlags.eventlogFlushTicks =
             RtsFlags.TraceFlags.eventlogFlushTime /
             RtsFlags.MiscFlags.tickInterval;
@@ -1879,6 +1890,9 @@ static void normaliseRtsOpts (void)
     if (RtsFlags.GcFlags.maxHeapSize != 0 &&
         RtsFlags.GcFlags.heapSizeSuggestion >
         RtsFlags.GcFlags.maxHeapSize) {
+        errorBelch("Maximum heap size (-M) is smaller than suggested heap size (-H)\n"
+                   "Setting maximum heap size to suggested heap size ( %" FMT_Word64 " )",
+                   (StgWord64) RtsFlags.GcFlags.maxHeapSize * (StgWord64) BLOCK_SIZE);
         RtsFlags.GcFlags.maxHeapSize = RtsFlags.GcFlags.heapSizeSuggestion;
     }
 
@@ -2159,6 +2173,10 @@ static void read_debug_flags(const char* arg)
         case 'l':
             RtsFlags.DebugFlags.linker = true;
             break;
+        case 'L':
+            RtsFlags.DebugFlags.linker_verbose = true;
+            RtsFlags.DebugFlags.linker = true;
+            break;
         case 'a':
             RtsFlags.DebugFlags.apply = true;
             break;
@@ -2177,6 +2195,9 @@ static void read_debug_flags(const char* arg)
         case 'C':
             RtsFlags.DebugFlags.compact = true;
             break;
+        case 'k':
+            RtsFlags.DebugFlags.continuation = true;
+            break;
         default:
             bad_option( arg );
         }
@@ -2202,7 +2223,7 @@ static bool read_heap_profiling_flag(const char *arg)
     bool error = false;
     switch (arg[2]) {
     case '\0':
-      errorBelch("-h is deprecated and will be removed in the next release (9.4),\nuse -hc instead.");
+      errorBelch("-h is deprecated, use -hc instead.");
       FALLTHROUGH;
     case 'C':
     case 'c':
@@ -2263,7 +2284,7 @@ static bool read_heap_profiling_flag(const char *arg)
                     RtsFlags.ProfFlags.bioSelector = selector;
                     break;
                 default:
-                    free(selector);
+                    stgFree(selector);
                 }
             }
             break;
@@ -2361,6 +2382,10 @@ static void read_trace_flags(const char *arg)
             RtsFlags.TraceFlags.sparks_sampled = enabled;
             RtsFlags.TraceFlags.sparks_full    = enabled;
             RtsFlags.TraceFlags.user           = enabled;
+            RtsFlags.TraceFlags.nonmoving_gc   = enabled;
+#if defined(TICKY_TICKY)
+            RtsFlags.TraceFlags.ticky          = enabled;
+#endif
             enabled = true;
             break;
 
@@ -2409,7 +2434,7 @@ static void read_trace_flags(const char *arg)
 }
 #endif
 
-static void GNU_ATTRIBUTE(__noreturn__)
+static void STG_NORETURN
 bad_option(const char *s)
 {
   errorBelch("bad RTS option: %s", s);
@@ -2594,14 +2619,14 @@ void freeRtsArgs(void)
 
 /*
 Note [OPTION_SAFE vs OPTION_UNSAFE]
-
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 Ticket #3910 originally pointed out that the RTS options are a potential
 security problem. For example the -t -s or -S flags can be used to
 overwrite files. This would be bad in the context of CGI scripts or
 setuid binaries. So we introduced a system where +RTS processing is more
 or less disabled unless you pass the -rtsopts flag at link time.
 
-This scheme is safe enough but it also really annoyes users. They have
+This scheme is safe enough but it also really annoys users. They have
 to use -rtsopts in many circumstances: with -threaded to use -N, with
 -eventlog to use -l, with -prof to use any of the profiling flags. Many
 users just set -rtsopts globally or in project .cabal files. Apart from
